#include "main.decl.h"
#include "main.h"
#include "multiply.decl.h"

CProxy_Main mainProxy;

Main::Main(CkArgMsg* msg) {
	int i,j,k;
	readyCount = 0;
	doneCount = 0;
	matrix_size = 0;
	
	if (msg->argc > 1){
		//Takes the matrix size from user input
		matrix_size = atoi(msg->argv[1]);
	}
	
	//Defines the block size
	int BLOCK_SIZE = matrix_size*matrix_size/CkNumPes();
	//Matrix X
	double (*matrix_x)[matrix_size] = (double (*)[matrix_size])new double[matrix_size*matrix_size];
	//Matrix Y
	double (*matrix_y)[matrix_size] = (double (*)[matrix_size])new double[matrix_size*matrix_size];
	//Holds the split for matrix X
	double (*split_x)[BLOCK_SIZE] = (double (*)[BLOCK_SIZE]) new double[matrix_size*matrix_size];
	//For future use
	split_bycol = new double[matrix_size*matrix_size];
	//Holds the split for matrix y
	double (*split_y)[BLOCK_SIZE] = (double (*)[BLOCK_SIZE]) split_bycol;	
	result = new double[matrix_size*matrix_size];
	//If the programm is runned on 1 processor, here will be kept the whole result
	double (*solution)[matrix_size] = (double (*)[matrix_size])new double[matrix_size*matrix_size];
 
	//Makes matrix X
	for (i = 0; i < matrix_size; i++) {
		for (j = 0; j < matrix_size; j++) {		
			matrix_x[i][j] = i*matrix_size+j;
		}
	}
	
	//Makes matrix Y
	for (i = 0; i < matrix_size; i++) {
		for (j = 0; j < matrix_size; j++) {		
			matrix_y[i][j] = i*matrix_size+j;
		}
	}
	
	//Prints for the user
	for (i = 0; i < matrix_size; i++) {
		CkPrintf("");
		for (j = 0; j < matrix_size; j++) {
			CkPrintf(" %.0f ",matrix_x[i][j]);			
		}
		CkPrintf("\n");
	}
	CkPrintf("\n");
	CkPrintf("   X   ");
	CkPrintf("\n");
	CkPrintf("\n");
	for (i = 0; i < matrix_size; i++) {
		CkPrintf("");
		for (j = 0; j < matrix_size; j++) {
			CkPrintf(" %.0f ",matrix_y[i][j]);			
		}
		CkPrintf("\n");
	}
	CkPrintf("\n");	

	//Transponent of matrix y
	for (i = 0; i < matrix_size; i++) {
		for (j = 0; j < matrix_size; j++) {	
			//bacause X and Y are equal
			matrix_y[j][i] = matrix_x[i][j];
		}
	}
	delete msg;

	//Splits the matricies into equal blocks, depending on the amout of processors used
	for (i = 0; i < CkNumPes(); i++) {
		for (j = 0; j < matrix_size/CkNumPes(); j++) {
			for (k = 0; k < matrix_size;k++){
				split_x[i][j*matrix_size + k] = matrix_x[i * matrix_size/CkNumPes() + j][k];
				split_y[i][j*matrix_size + k] = matrix_y[i * matrix_size/CkNumPes() + j][k];
			}		
		}
	}


	CkPrintf("Running with %d elements using %d processors.\n",matrix_size, CkNumPes());
	mainProxy = thisProxy;
	//Makes chare object array
	object_array = CProxy_Multiply::ckNew(CkNumPes());
	//If we use only one processor
	if (CkNumPes() == 1){
		for (i = 0; i < matrix_size; i++) {
			for (j = 0; j < matrix_size; j++) {
				for (k = 0; k < matrix_size;k++){
					solution[i][j] += matrix_x[i][k] * matrix_y[j][k];
				}	
			}
		}		
	}
	else{
		//Sends the slices/rows to be stored
		for (i = 0; i < CkNumPes(); i++) {	  
			object_array[i].store(matrix_size * matrix_size/CkNumPes(), split_x[i],matrix_size);	  
		}
	}
}

Main::Main(CkMigrateMessage* msg) { }

//Checks if all rows are saved for future collisions
void Main::ready()  {
	int i;
	readyCount += 1;
	int BLOCK_SIZE = matrix_size*matrix_size/CkNumPes();
	double (*split_y)[BLOCK_SIZE] = (double (*)[BLOCK_SIZE]) split_bycol;

	if (readyCount == CkNumPes()) {
		CkPrintf("Start calculation\n");
		//Sends the slices for parallel computing
		for (i = 0; i < CkNumPes(); i++) {	  
			object_array[i].multiply(matrix_size * matrix_size/CkNumPes(), split_y[i],i);	  
		}
	}
}

void Main::save_temp(int SIZE, double *number, int row_number){
	int i,j;
	double (*temp)[matrix_size] = (double (*)[matrix_size]) result;
	
	//Stores gathered results
	for(i = 0; i < matrix_size/CkNumPes(); i++){
		for(j = 0; j < matrix_size; j++) {
			temp[i+row_number*matrix_size/CkNumPes()][j] = number[i*matrix_size+j];
		}
	}
	
 	doneCount ++;
	if (doneCount == CkNumPes()){
		CkPrintf("\n");
		CkPrintf("Tulemus: \n");
		for (i = 0; i < matrix_size; i++) {			
			for (j = 0; j < matrix_size; j++) {			
	  			CkPrintf(" %.0f ",temp[i][j]);					
			}
			CkPrintf("\n");
		}
		CkExit();
	}
}

#include "main.def.h"