#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

#define SIZE 4
#define P 4

int main(int argc, char** argv) {

	MPI_Init(NULL, NULL);

	// Get the number of processes
	int world_size;
	MPI_Comm_size(MPI_COMM_WORLD, &world_size);

	// Get the rank of the process
	int world_rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);

	// Get the name of the processor
	char processor_name[MPI_MAX_PROCESSOR_NAME];
	int name_len;
	MPI_Get_processor_name(processor_name, &name_len);

	//Every process saves a block of result
	int (*temp)[SIZE] = malloc((SIZE / P)*(SIZE)*sizeof(int));
	//Matrix X
	int (*x)[SIZE] = malloc(SIZE*SIZE*sizeof(int));
	//Matrix Y
	int (*y)[SIZE] = malloc(SIZE*SIZE*sizeof(int));
	//Transponent of matrix Y
	int (*y_t)[SIZE] = malloc(SIZE*SIZE*sizeof(int));
	//Block of matrix X
	int *row = malloc((SIZE / P)*(SIZE)*sizeof(int));
	//Block of matrix Y
	int *column = malloc((SIZE / P)*(SIZE)*sizeof(int));
	int *column_temp = malloc((SIZE / P)*(SIZE)*sizeof(int));
	int counter = 0;

	int i,j,k,p;
	MPI_Request request;

	if (world_rank == 0){

		for (i = 0; i < SIZE; i++){
			for (j = 0; j < SIZE; j++){
    			x[i][j] = SIZE*i+j;
  
			}
		}
				
		for (i = 0; i < SIZE; i++){
			for (j = 0; j < SIZE; j++){
    			y[i][j] = SIZE*i+j;
  
			}
		}

		for (i = 0; i < SIZE; i++){
			for (j = 0; j < SIZE; j++){
    			printf("%d ", x[i][j]);  
			}
			printf("\n"); 
		}

		printf("\n"); 
		printf("   x");
		printf("\n");
		printf("\n");
				
		for (i = 0; i < SIZE; i++){
			for (j = 0; j < SIZE; j++){
    			printf("%d ", y[i][j]);  
			}
			printf("\n"); 
		}

		//Makes matrix Y transponent		
		for (i = 0; i < SIZE; i++){
			for (j = 0; j < SIZE; j++){
    			y_t[i][j] = y[j][i];
  
			}
		}
	}

	//Splits matrix X and transponent matrix Y into equal blocks and sends to every processor available
	MPI_Scatter(x, SIZE * SIZE/P, MPI_INT,  row, SIZE * SIZE/P, MPI_INT, 0,    MPI_COMM_WORLD);
	MPI_Scatter(y_t, SIZE * SIZE/P, MPI_INT,  column, SIZE * SIZE/P, MPI_INT, 0,    MPI_COMM_WORLD);

	//Every processor calculates given blocks
	for (i = 0; i < SIZE/P; i++){
		for (j = 0; j < SIZE/P; j++){			
			temp[i][SIZE/P * world_rank + j] = 0;
			for (k= 0; k < SIZE; k++){
				temp[i][SIZE/P * world_rank + j] += row[i*SIZE + k] * column[j*SIZE + k];		
			}
		}		
	}

	for (p= 1; p < P; p++){
		//Every processor sends his column to neighbour asynchronously
		MPI_Isend(column, SIZE * SIZE/P, MPI_INT, (world_rank-1+P)%P, 5, MPI_COMM_WORLD, &request);			
		//Every processor receives the column block to further calculation
		MPI_Recv(column_temp,SIZE * SIZE/P, MPI_INT,(world_rank +1)%P ,5,MPI_COMM_WORLD,MPI_STATUS_IGNORE);	
			//Every process makes calculation by multiplying processors own row block and received bolumn block
			for (i = 0; i < SIZE/P; i++){
				for (j = 0; j < SIZE/P; j++){	
					temp[i][SIZE/P * ((world_rank+p)%P) + j] = 0;
					for (k= 0; k < SIZE; k++){
						temp[i][SIZE/P * ((world_rank+p)%P) + j]  +=  row[i*SIZE + k] * column_temp[j*SIZE + k];		
					}
				}
			}
		//For avoiding collisions
		MPI_Wait(&request, MPI_STATUS_IGNORE);
		for (i = 0; i < SIZE*SIZE/P; i++){
			column[i] = column_temp[i];
 		}
	}	

	int result[SIZE*SIZE];
	//Gathers the results
	MPI_Gather(temp, SIZE*SIZE/P, MPI_INT, result, SIZE*SIZE/P, MPI_INT,0,MPI_COMM_WORLD);

	//Only main processor should print
	if (world_rank == 0){
		printf("\n"); 
		printf("Tulemus: \n");
		for (i = 0; i < SIZE; i++){
			for (j = 0; j < SIZE; j++){
    			printf("%d ", result[i*SIZE+j]);  
			}
			printf("\n"); 
		}

	}  
	// Finalize the MPI environment. No more MPI calls can be made after this
	MPI_Finalize();
}
